#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 14:23:57 2021

@author: nahuel
"""

import scipy.integrate
import numpy as np

from ResNova.GenericSource import GenericSource
from ResNova.Target import CompositeTarget

class Detector:
    def __init__(self, name: str, 
                 model, 
                 target: CompositeTarget, 
                 mass = 1000.,
                 energy_resolution = 0.2,
                 energy_threshold = 1.,
                 time_resolution = 200e-3):
        '''
        

        Parameters
        ----------
        name : str
            Name of the detector
        model : list of GenericSource
            List of models.
        target : CompositeTarget
            Target material.
        mass : mass, optional
            detectors mass in kg. The default is 1000..
        energy_resolution : float, optional
            nuclear recoil energy resolution in keV. The default is 0.2.
        energy_threshold : float, optional
            nuclear recoil energy threshold in keV. The default is 1.

        Returns
        -------
        None.

        '''
        self.name = name
        self.model = model
        self.mass = mass
        self.target = target
        self.energy_resolution = energy_resolution
        self.energy_threshold = energy_threshold
        self.time_resolution = time_resolution
    
    def recoil_spectrum(self, Er: float, t: float = None) -> float:
        '''
        

        Parameters
        ----------
        Er : float
            Recoil energy in GeV.
        t: float
            Time at which the flux is evaluated
        Returns
        -------
        dictionary
            counts/GeV/s for the given detector mass per model and the sum
            with key name 'sum'.

        '''

        allrates = {}

        for mod in self.model:
            rate_per_element = np.zeros_like(self.target.elements)
            for idx, El in enumerate(self.target.elements):
                rate_per_element[idx] = self.mass*self.target.relative_mass[El.name]*1000/El.A*\
                    6.02214076e+23*\
                np.vectorize(
                    lambda y: scipy.integrate.quad(
                        lambda x: mod.flux(x, t=t)*1e-4*\
                            El.dSigmaCNNSdEr(Er = y, Enu = x),
                        El.Enumin(y), 0.05)[0]
                    )(Er)
            allrates[mod.name] = np.sum(rate_per_element)
        allrates['sum'] = 0
        for mod in self.model:
            allrates['sum'] += allrates[mod.name]
        return allrates
    
    def recoil_rate(self, t: float = None) -> float:
        '''
        

        Parameters
        ----------
        t : float
            time in s.
        Returns
        -------
        dictionary
            counts/s for the given detector mass.

        '''

        allrates = {}

        for mod in self.model:
            cr = np.zeros_like(self.target.elements)
            for idx, El in enumerate(self.target.elements):
                cr[idx] = self.mass*self.target.relative_mass[El.name]*1000/El.A*\
                    6.02214076e+23*\
                np.vectorize(
                    lambda z: scipy.integrate.nquad(
                        lambda x, y: mod.flux(Enu = x, t = z)*1e-4*\
                            El.dSigmaCNNSdEr(Er = y, Enu = x),
                        [ [0,0.10], [self.energy_threshold*1e-6, 20e-6]])[0]
                    )(t)
            allrates[mod.name] = np.sum(cr) 
        allrates['sum'] = 0
        for mod in self.model:
            allrates['sum'] += allrates[mod.name]
        return allrates
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 00:43:13 2021

@author: nahuel
"""

import numpy as np
#fermi constant [GeV^-2];
GF = 1.1663787e-5 

def WeakCharge(A: float, Z: float) -> float:
    '''
    

    Parameters
    ----------
    A : float
        Atomic mass number.
    Z : int
        Atomic number.

    Returns
    -------
    float
        Absolute value of the nuclear weak charge.

    '''
    N = float(A) - float(Z)
    return np.abs(float(N)-0.0599*float(Z))

class TargetElement:
    def __init__(self, A: float, Z: float, name: str, s = 0.9):
        '''
        

        Parameters
        ----------
        A : float
            Atomic mass number of the target element (amu).
        Z : int
            Atomic number of the target element.
        name: str
            Name of the species
        s : float
            Nuclear skin thickness [fm].

        Returns
        -------
        None.

        '''
        self.A = A
        self.Z = Z
        self.Qw = WeakCharge(A, Z)
        self.s = s
        self.Mn = 0.9315*float(A)#931.5*1e3*float(A)
        self.name = name
    
    def Helm_Form_Factor2(self, Er: float):
        '''
        

        Parameters
        ----------
        Er : float
            Recoil energy [GeV].

        Returns
        -------
        TYPE
            Helm form factor squared at energy Er.

        '''
        amu = 931.5*1e+9
        A = self.A
        #Convert recoil energy to momentum transfer q in keV
        q1 = np.sqrt(2*A*amu*Er)
    
        #Convert q into fm^-1
        q2 = q1*(1e-12/1.97e-7)
    
        #Calculate nuclear parameters
        s = self.s
        a = 0.52
        c = 1.23*(A**(1.0/3.0)) - 0.60
        R1 = np.sqrt(c*c + 7*np.pi*np.pi*a*a/3.0 - 5*s*s)

        #Calculate form factor
        x = q2*R1
        J1 = np.sin(x)/(x*x) - np.cos(x)/x
        F = 3*J1/x
     
        return (F*F)*(np.exp(-(q2*q2*s*s)))
    

    def Ermax(self, Enu: float):
        '''
        

        Parameters
        ----------
        Enu : float
            Energy of the incoming neutrino [GeV].

        Returns
        -------
        float
            Maximum recoil energy given Enu [GeV].

        '''
        return 2*Enu**2/(self.Mn)

    def Enumin(self, Er: float):
        '''
        

        Parameters
        ----------
        Er : float
            Recoil energy [GeV].

        Returns
        -------
        float
            Minimum incoming neutrino energy needed to induce a recoil
            of energy Enumin [GeV].

        '''
        return np.sqrt(self.Mn*Er/2)
    
    def dSigmaCNNSdEr(self, Er: float, Enu: float) -> float:
        '''
        

        Parameters
        ----------
        Er : float
            Recoiling energy [GeV].
        Enu : float
            Incident neutrini energy [GeV].

        Returns
        -------
        float
            cross-section [cm^2/GeV].

        '''
        #GeV^-2 -> cm^2, Fermi, weakcharge, mass number, kinematics, nuclear form
        return 389.4*1e-30*GF**2/(4.*np.pi)*self.Qw**2*self.Mn*\
        (1.-np.divide(
            self.Mn*Er,(2*Enu**2), 
            out = np.zeros_like(Er), where=Er!=0))*\
            self.Helm_Form_Factor2(Er)*(Er <= self.Ermax(Enu))
class CompositeTarget:
    def __init__(self, elements, subscripts):
        '''
        

        Parameters
        ----------
        elements : list of TargetElement
            List containing TargetElements to be considered in the composite
            material.
        subscript : list of ints
            List of molecular subscripts fo each element, indicating the
            number of each species present in one molecule.


        Returns
        -------
        None.

        '''
        if any(type(i) != type(TargetElement(0, 0, '')) for i in elements):
            raise TypeError('elements must be a list of TargetElements')
        if any(type(i) != int 
               for i in subscripts):
            raise TypeError('subscripts must be a list of\
                            ints')
        if len(elements) != len(subscripts):
            raise ValueError('lengths of elements and subscripts \
                             must be equal')                  
        self.elements = elements
        self.subscripts = subscripts
        self.relative_mass = {}
        tot = 0.
        for i in range(len(elements)):
            tot += elements[i].A*subscripts[i]
        #molar mass in kg
        self.molar_mass = 1e-3*tot
        for idx, i in enumerate(elements):
            self.relative_mass[i.name] = i.A*subscripts[idx]/tot 
        
TargetPb = TargetElement(A = 208, Z = 82, name = 'Pb')
TargetCa = TargetElement(A = 40, Z = 20, name = 'Ca')
TargetW = TargetElement(A = 183.8, Z = 74, name = 'W')
TargetO = TargetElement(A = 16, Z = 8, name = 'O')
TargetXe = TargetElement(A = 131.3, Z = 54, name = 'Xe')

PbWO4 = CompositeTarget([TargetPb, TargetW, TargetO], [1, 1, 4])
CaWO4 = CompositeTarget([TargetCa, TargetW, TargetO], [1, 1, 4])
MetallicPb = CompositeTarget([TargetPb], [1])
Xe = CompositeTarget([TargetXe], [1])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 16:31:16 2021

@author: nahuel
"""

class GenericSource:
    def __init__(self, name):
        self.name = name
        pass
    
    def flux(self, Enu: float, t: float = None) -> float:
        '''
        

        Parameters
        ----------
        Enu : float
            Particle energy in GeV.
        t: float
            Time in s
            
        Returns
        -------
        float
            Flux of neutrinos 
            in particles/GeV/m^2/s.

        '''
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug  1 23:29:05 2021
@author: nahuel
"""

"""
Abstract class defining a model:
    it must offer access to the relevant physical variables:
        luminosity [foe/s], fluence [MeV/s], Eavg [MeV], E2avg [MeV^2], alpha
"""

import numpy as np
import scipy.integrate
from scipy.special import gamma
from ResNova.GenericSource import GenericSource

class SNModel(GenericSource):
    def __init__(self, name, distance = 10.):
        '''
        

        Parameters
        ----------
        name : str
            Name of the model.
        distance : float, optional
            SN distance in kpc. The default is 10..

        Returns
        -------
        None.

        '''
        self.name = name
        #distance from kpc to m
        self.distance = distance*3.08567758e+19

    def luminosity(self, time: float) -> float:
        '''
        

        Parameters
        ----------
        time : float
            Time in seconds at which the luminosity is returned.

        Returns
        -------
        float
            Luminosity in foe/s.

        '''
        pass
    
    def fluence(self, time: float) -> float:
        '''
        

        Parameters
        ----------
        time : float
            Time in seconds at which the fluence is returned.

        Returns
        -------
        float
            fluence in GeV/s.

        '''
        pass
    
    def Eavg(self, time: float) -> float:
        '''
        

        Parameters
        ----------
        time : float
            Time in seconds at which the average neutrino energy is returned.

        Returns
        -------
        float
            average neutrino energy in MeV.

        '''
        pass
    
    def E2avg(self, time: float) -> float:
        '''
        

        Parameters
        ----------
        time : float
            Time in seconds at which the quadratic average 
            neutrino energy is returned.

        Returns
        -------
        float
            quadratic average neutrino energy in MeV^2.

        '''
        pass
    
    def alpha(self, time: float) -> float:
        '''
        

        Parameters
        ----------
        time : float
            Time in seconds at which the pinching parameter is returned.

        Returns
        -------
        float
            pinching parameter.

        '''
        pass
    
    def SetTimeInterval(self, t = (float, float)):
        '''
        

        Parameters
        ----------
        t : (float, float)
            time interval to consider the SN as a static source
            (astrophysical quantities are averaged over the interval)

        Returns
        -------
        None.

        '''
        self.alpham = scipy.integrate.quad(
                self.alpha, t[0], t[1]
                )[0]/(t[1] - t[0])
        self.Eavgm = scipy.integrate.quad(
                self.Eavg, t[0], t[1]
                )[0]/(t[1] - t[0])    
        self.Lm = scipy.integrate.quad(
                self.fluence, t[0], t[1]
                )[0]/(t[1] - t[0]) 
        
    def phi(self, Enu: float, t: float) -> float:
        '''
        

        Parameters
        ----------
        Enu : float
            Neutrino energy in GeV.
        t: float
            Time is seconds at which the distribution is computed. If t is None, static
            values will be used

        Returns
        -------
        float
            Distribution function of emitted neutrinos, unit normalized.

        '''
        if t == None:
            alpha = self.alpham
            Em = self.Eavgm
        else:
            alpha = self.alpha(t)
            Em = self.Eavg(t)
            
        norm = gamma(alpha + 1)/((alpha+1)/Em)**(alpha+1)/Em**alpha
        return 1/norm*(Enu/Em)**alpha*np.exp(-(alpha+1)*Enu/Em)
    
    def flux(self, Enu: float, t: float = None):
        '''
        

        Parameters
        ----------
        Enu : float
            Neutrino energy in GeV.
        t : float 
            Time is seconds at which the flux is computed.
            
        Returns
        -------
        float
            Flux of neutrinos of given energy at distance d, at time t
            in neutrinos/GeV/m^2/s.

        '''
        
        if t == None:
            L = self.Lm
            Eavg = self.Eavgm
        else:
            L = self.fluence(t)
            Eavg = self.Eavg(t)
        
            
        return L/4/np.pi/(self.distance)**2/Eavg*\
            self.phi(Enu, t)
    


# -*- coding: utf-8 -*-
from setuptools import setup

setup(
   name='ResNova',
   version='0.1.0',
   author='Nahuel Ferreiro',
   author_email='ferreiro@mpp.mpg.de',
   packages=['ResNova'],
   license='LICENSE.txt',
   description='Cryogenic detectors for neutrino detection',
   long_description='For now no long description',
   url = 'https://github.com/notacustomdev/ResNova',
   download_url = 'https://github.com/notacustomdev/ResNova/archive/refs/tags/v_010.tar.gz',
   #long_description=open('README.txt').read(),
   install_requires=[
       "numpy>=1.19.2",
       "scipy>=1.5.2",
       "pandas",
       "matplotlib"
   ],
)
